/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Josh Aas <josh@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "BasicPlugin.h"
#include "nppalmdefs.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>

#define PLUGIN_NAME        "Basic Sample Plug-in"
#define PLUGIN_DESCRIPTION PLUGIN_NAME " (Mozilla SDK)"
#define PLUGIN_VERSION     "1.0.0.0"
#define PLUGIN_MIMEDESCRIPTION "application/x-webosinternals-basicplugin:bsc:Basic plugin"

#define max(a, b)		(	((a) >= (b))	? (a) : (b)	)
#define min(a, b)		(	((a) <= (b))	? (a) : (b)	)
#define RGB(r, g, b)	(colorAlpha | ((r)<<16) | ((g)<<8) | (b))

typedef enum
	{
	colorAlpha		= 0xFF000000,
	colorDistBlue	= 0xaabbccdd,
	colorBlack		= colorAlpha | 0,
	colorWhite		= colorAlpha | 0xFFFFFF,
	colorLightGrey	= colorAlpha | 0x808080,
	colorMediumGrey	= colorAlpha | 0x777777,
	colorRed		= colorAlpha | 0xFF0000,
	colorGreen		= colorAlpha | 0x00FF00,
	colorBlue		= colorAlpha | 0x0000FF,
	colorFuchsia	= colorAlpha | 0xFF00FF,
	colorYellow		= colorAlpha | 0xFFFF00,
	colorAqua		= colorAlpha | 0x00FFFF,
	}eColor;


static NPNetscapeFuncs NPNFuncs;

struct MyObject : NPObject {
  NPP instance;
  int event_x;
  int event_y;
  int event_key_code;
  int solid_lines;
  MyObject(NPP i) : instance(i), event_x(0), event_y(0), event_key_code(0), solid_lines(0) { }
};

struct InstanceData {
//  NPP npp;
//  NPWindow window;
  MyObject *object;
  InstanceData() : object(0) { }
};

int setSize(NPP instance, int width, int height);
void logmsgv(const char* desc, NPPVariable variable)
{
    fprintf(stderr, "basic: %s - %s (#%i)\n", desc, varname(variable), variable);
}

void logTimeStamp(FILE *out, int line)
	{
	struct tm *t;
	time_t now;

	if(!out)
		return;
	time(&now);
	t = localtime(&now);

	fprintf(out, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d	%d	basic: ",
			t->tm_year + 1900,
			t->tm_mon + 1,
			t->tm_mday,
			t->tm_hour,
			t->tm_min,
			t->tm_sec, line);
	}

#define debugLogLine(x)								debugLog(__LINE__, x)
static void debugLogArgv(int line, int argc, char **argn, char **argv, const char *msg)
	{
	int j;
	FILE *out = stderr;//fopen("/tmp/basicPluginOutput-out.log", "a");
	const char *prefix = "args =";
	if(!out)
		return;
	logTimeStamp(out, line);
	fprintf(out, "%s(", msg);
	for(j = 0; argc > 0; argc--, argv++, argn++, j++, prefix = ",")
		fprintf(out, "%s %d:%s[%s]", prefix, j, *argn, *argv);
	fprintf(out, ")\n");
	if(out != stderr)
		fclose(out);
	}

static void debugLog(int line, const char *format, ...)
	{
	FILE *out = stderr;//fopen("/tmp/basicPluginOutput-out.log", "a");
	if(!out)
		return;
	logTimeStamp(out, line);
	va_list(arglist);
	va_start(arglist, format);
	vfprintf(out, format, arglist);
	fprintf(out, "\n");
	if(out != stderr)
		fclose(out);
	}

void debugLogVariant(int line, NPVariant *var, bool ok, const char *msg)
	{
	if(ok)
		{
		if(NPVARIANT_IS_INT32(*var))
			debugLog(line, "%s = %d val:%d", msg, ok, NPVARIANT_TO_INT32(*var));
		else if(NPVARIANT_IS_DOUBLE(*var))
			debugLog(line, "%s = %d val:%f", msg, ok, NPVARIANT_TO_DOUBLE(*var));
		else if(NPVARIANT_IS_BOOLEAN(*var))
			debugLog(line, "%s = %d val:%s", msg, ok, NPVARIANT_TO_BOOLEAN(*var) ? "true" : "false");
		else if(NPVARIANT_IS_STRING(*var))
			debugLog(line, "%s = %d val:[%s]", msg, ok, NPVARIANT_TO_STRING(*var).UTF8Characters);
		else if(NPVARIANT_IS_OBJECT(*var))
			debugLog(line, "%s = %d val:%p", msg, ok, NPVARIANT_TO_OBJECT(*var));
		else
			debugLog(line, "%s = %d type:%d", msg, ok, var->type);
		}
	else
		debugLog(line, "%s = %d", msg, ok);
	}

static void
fillPluginFunctionTable(NPPluginFuncs* pFuncs)
{
  pFuncs->version = 11;
  pFuncs->size = sizeof(*pFuncs);
  pFuncs->newp = NPP_New;
  pFuncs->destroy = NPP_Destroy;
  pFuncs->setwindow = NPP_SetWindow;
  pFuncs->newstream = NPP_NewStream;
  pFuncs->destroystream = NPP_DestroyStream;
  pFuncs->asfile = NPP_StreamAsFile;
  pFuncs->writeready = NPP_WriteReady;
  pFuncs->write = NPP_Write;
  pFuncs->print = NPP_Print;
  pFuncs->event = NPP_HandleEvent;
  pFuncs->urlnotify = NPP_URLNotify;
  pFuncs->getvalue = NPP_GetValue;
  pFuncs->setvalue = NPP_SetValue;
}

NP_EXPORT(NPError)
NP_Initialize(NPNetscapeFuncs* bFuncs, NPPluginFuncs* pFuncs)
{
  debugLogLine("NP_Initialize");

  NPNFuncs = *bFuncs;

  fillPluginFunctionTable(pFuncs);

  return NPERR_NO_ERROR;
}

NP_EXPORT(char*)
NP_GetPluginVersion()
{
  debugLogLine("NP_GetPluginVersion");
  return (char *)PLUGIN_VERSION;
}

NP_EXPORT(char*)
NP_GetMIMEDescription()
{
  debugLogLine("NP_GetMIMEDescription");
  return (char *)PLUGIN_MIMEDESCRIPTION;
}

NPObject *NPN_CreateObject(NPP npp, NPClass *aClass)
{
  return NPNFuncs.createobject(npp, aClass);
}

void NPN_ForceRedraw(NPP npp)
{
  return NPNFuncs.forceredraw(npp);
}

bool NPN_IdentifierIsString(NPIdentifier identifier)
{
  return NPNFuncs.identifierisstring(identifier);
}

NPUTF8 *NPN_UTF8FromIdentifier(NPIdentifier identifier)
{
  return NPNFuncs.utf8fromidentifier(identifier);
}

static NPObject *Object_Allocate(NPP npp,NPClass *aClass)
{
  return new MyObject(npp);
}

static void Object_Deallocate(NPObject *npobj)
{
  delete (MyObject *)npobj;
}

static bool
  Object_Construct(
    NPObject *npobj,
    const NPVariant *args,
    uint32_t argCount,
    NPVariant *result
  )
{
  debugLogLine("In Object_Construct");
  return false;
}

static bool ContainsName(const char **name_array,NPIdentifier name)
{
  bool is_string = NPN_IdentifierIsString(name);
  if (!is_string) {
    debugLogLine("  Identifier is not a string.");
    return false;
  }
  const char *name_str = NPN_UTF8FromIdentifier(name);
  debugLog(__LINE__, "  Identifier = %s",name_str);
  int index = 0;
  for (;;++index) {
    const char *one_name = name_array[index];
    if (!one_name) {
      break;
    }
    if (strcmp(name_str,one_name)==0) {
      return true;
    }
  }
  return false;
}


static bool Object_HasProperty(NPObject *npobj, NPIdentifier name)
{
  debugLogLine("In Object_HasProperty");
  static const char *property_names[] = {
    0
  };
  return ContainsName(property_names,name);
}


static bool Object_HasMethod(NPObject *npobj,NPIdentifier name)
{
  debugLogLine("In Object_HasMethod");
  static const char *method_names[] = {
    "sendTap","sendKey",0
  };
  return ContainsName(method_names,name);
}


static void PrintValue(const NPVariant *arg)
{
  if (NPVARIANT_IS_VOID(*arg)) {
    fprintf(stderr,"void");
  }
  else if (NPVARIANT_IS_NULL(*arg)) {
    fprintf(stderr,"null");
  }
  else if (NPVARIANT_IS_BOOLEAN(*arg)) {
    if (NPVARIANT_TO_BOOLEAN(*arg)) {
      fprintf(stderr,"true");
    }
    else {
      fprintf(stderr,"false");
    }
  }
  else if (NPVARIANT_IS_INT32(*arg)) {
    fprintf(stderr,"%d",NPVARIANT_TO_INT32(*arg));
  }
  else if (NPVARIANT_IS_DOUBLE(*arg)) {
    fprintf(stderr,"%g",NPVARIANT_TO_DOUBLE(*arg));
  }
  else if (NPVARIANT_IS_STRING(*arg)) {
    NPString value = NPVARIANT_TO_STRING(*arg);
    const NPUTF8 *chars = value.UTF8Characters;
    uint32_t len = value.UTF8Length;
    fprintf(stderr,"\"");
    {
      uint32_t i = 0;
      for (;i!=len;++i) {
        fprintf(stderr,"%c",chars[i]);
      }
    }
    fprintf(stderr,"\"");
  }
  else if (NPVARIANT_IS_OBJECT(*arg)) {
    fprintf(stderr,"object");
  }
  else {
    fprintf(stderr,"unknown");
  }
}


static void PrintArgs(const NPVariant *args,int argCount)
{
  fprintf(stderr,"(");
  int argnum = 0;
  for (;argnum!=argCount;++argnum) {
    if (argnum!=0) {
      fprintf(stderr,",");
    }
    PrintValue(&args[argnum]);
  }
  fprintf(stderr,")");
}


static int IntValue(const NPVariant *value)
{
  if (NPVARIANT_IS_INT32(*value)) {
    int32_t value_int32 = NPVARIANT_TO_INT32(*value);
    return value_int32;
  }
  else if (NPVARIANT_IS_DOUBLE(*value)) {
    double value_double = NPVARIANT_TO_DOUBLE(*value);
    return (int)value_double;
  }
  else if (NPVARIANT_IS_BOOLEAN(*value)) {
    bool value_bool = NPVARIANT_TO_BOOLEAN(*value);
    return (int)value_bool;
  }
  else if (NPVARIANT_IS_STRING(*value)) {
    const NPString &value_str = NPVARIANT_TO_STRING(*value);
    const char *value_chars = value_str.UTF8Characters;
    debugLog(__LINE__, "String is %s",value_chars);
    return (int)value_chars[0];
  }
  else {
    debugLog(__LINE__, "Expecting int, but got type %d",value->type);
    return 0;
  }
}

void HandleKeyPress(MyObject *myobj)
	{
	if(myobj->event_key_code == 'S')
		myobj->solid_lines = 1;
	else if(myobj->event_key_code == 'C')
		myobj->solid_lines = 0;
	else if(myobj->event_key_code == 'T')
		myobj->solid_lines = !myobj->solid_lines;
	else
		return;
	NPNFuncs.forceredraw(myobj->instance);
	debugLog(__LINE__, "HandleKeyPress(key:'%c' solid_lines:%d)", myobj->event_key_code, myobj->solid_lines);
	}

bool
  Object_Invoke(
    NPObject *npobj,
    NPIdentifier name,
    const NPVariant *args,
    uint32_t argCount,
    NPVariant *result
  )
{
  MyObject *myobj = (MyObject *)npobj;
  bool is_string = NPN_IdentifierIsString(name);
  if (!is_string) {
    debugLogLine("  Object_Invoke: Method name is not a string");
    return false;
  }
  const char *name_str = NPN_UTF8FromIdentifier(name);
  fprintf(stderr,"  Object_Invoke: %s",name_str);
  PrintArgs(args,argCount);
  fprintf(stderr,"\n");
  if (strcmp(name_str,"sendTap")==0) {
    if (argCount<2) {
      debugLogLine("  wrong number of args");
      return false;
    }
    myobj->event_x = IntValue(&args[0]);
    myobj->event_y = IntValue(&args[1]);
	debugLog(__LINE__, "Object_Invoke(sendTap: event_x:%d, event_y:%d)", myobj->event_x, myobj->event_y);
    NPN_ForceRedraw(myobj->instance);
    return true;
  }
  else if (strcmp(name_str,"sendKey")==0) {
    if (argCount<1) {
      debugLogLine("  wrong number of args");
      return false;
    }
    myobj->event_key_code = IntValue(&args[0]);
	HandleKeyPress(myobj);
	debugLog(__LINE__, "Object_Invoke(sendKey: event_key_code:%d)", myobj->event_key_code);
    return true;
  }

  return false;
}


static bool
  Object_SetProperty(
    NPObject *npobj,
    NPIdentifier name,
    const NPVariant *value
  )
{
  MyObject *myobj = (MyObject *)npobj;
  debugLogLine("In Object_SetProperty");
  bool is_string = NPN_IdentifierIsString(name);
  debugLog(__LINE__, "  IdentifierIsString = %d",is_string);
  if (is_string) {
    const char *name_str = NPN_UTF8FromIdentifier(name);
    debugLog(__LINE__, "  Identifier = %s",name_str);
    if (strcmp(name_str,"event_x")==0) {
      myobj->event_x = IntValue(value);
      debugLog(__LINE__, "  event_x set to %d",myobj->event_x);
    }
    else if (strcmp(name_str,"event_key_code")==0) {
      myobj->event_key_code = IntValue(value);
      debugLog(__LINE__, "  event_key_code set to %d",myobj->event_key_code);
    }
    else if (strcmp(name_str,"event_y")==0) {
      myobj->event_y = IntValue(value);
      debugLog(__LINE__, "  event_y set to %d",myobj->event_y);
    }
    else if (strcmp(name_str,"test_property")==0) {
      int command = IntValue(value);
      debugLog(__LINE__, "  command=%d",command);
      if (command==2) {
        NPN_ForceRedraw(myobj->instance);
      }
    }
    else {
      debugLogLine("Invalid property");
    }
  }
  return false;
}

static NPObject *GetScriptableObject(NPP instance)
{
  InstanceData *data = (InstanceData *)instance->pdata;
  if (!data->object) {
    static NPClass object_class;
    object_class.structVersion = NP_CLASS_STRUCT_VERSION;
    object_class.allocate = Object_Allocate;
    object_class.deallocate = Object_Deallocate;
    object_class.invalidate = 0;
    object_class.hasMethod = Object_HasMethod;
    object_class.invoke = Object_Invoke;
    object_class.invokeDefault = 0;
    object_class.hasProperty = Object_HasProperty;
    object_class.getProperty = 0;
    object_class.setProperty = Object_SetProperty;
    object_class.removeProperty = 0;
    object_class.enumerate = 0;
    object_class.construct = Object_Construct;

    data->object =
      (MyObject *)NPN_CreateObject(
        instance,
        &object_class
      );
  }

  return data->object;
}

NP_EXPORT(NPError)
NP_GetValue(NPP instance, NPPVariable aVariable, void* aValue) {
	debugLog(__LINE__, "NP_GetValue: [%s (#%i)]", varname(aVariable), aVariable);

    switch (aVariable) {
    case NPPVpluginNameString:
      *((const char**)aValue) = PLUGIN_NAME;
      break;
    case NPPVpluginDescriptionString:
      *((const char**)aValue) = PLUGIN_DESCRIPTION;
      break;
    default:
      return NPERR_INVALID_PARAM;
      break;
  }
  return NPERR_NO_ERROR;
}

NP_EXPORT(NPError)
NP_Shutdown()
{
  debugLogLine("NP_Shutdown");
  return NPERR_NO_ERROR;
}

NPError
NPP_New(NPMIMEType pluginType, NPP instance, uint16_t mode, int16_t argc, char* argn[], char* argv[], NPSavedData* saved) {
  instance->pdata = new InstanceData;
  debugLogArgv(__LINE__, argc, argn, argv, "NPP_New");
  debugLog(__LINE__, "NPP_New([%s], mode:%d, argc:%d, instance:[pdata:%p ndata:%p])", pluginType, mode, argc, instance->pdata, instance->ndata);

	//	XXX - We are changing the size of the Plugin area
  setSize(instance, /*width*/-1, /*height*/3000);
  return NPERR_NO_ERROR;
}

NPError
NPP_Destroy(NPP instance, NPSavedData** save) {
  debugLog(__LINE__, "NPP_Destroy(instance:[pdata:%p ndata:%p])", instance->pdata, instance->ndata);
  InstanceData* instanceData = (InstanceData*)(instance->pdata);
  if (instanceData) {
    delete instanceData;
  }
  return NPERR_NO_ERROR;
}

NPError
NPP_SetWindow(NPP instance, NPWindow* window) {
	if(window && window->window)
		debugLog(__LINE__, "NPP_SetWindow(x:%d, y:%d, width:%d, height:%d, type:%d)", window->x, window->y, window->width, window->height, window->type);
	else		//	Destroying window
		debugLog(__LINE__, "NPP_SetWindow(window:%p)", window);
  return NPERR_NO_ERROR;
}

NPError
NPP_NewStream(NPP instance, NPMIMEType type, NPStream* stream, NPBool seekable, uint16_t* stype) {
  debugLogLine("NPP_NewStream");
  return NPERR_GENERIC_ERROR;
}

NPError
NPP_DestroyStream(NPP instance, NPStream* stream, NPReason reason) {
  debugLogLine("NPP_DestroyStream");
  return NPERR_GENERIC_ERROR;
}

int32_t
NPP_WriteReady(NPP instance, NPStream* stream) {
  debugLogLine("NPP_WriteReady");
  return 0;
}

int32_t
NPP_Write(NPP instance, NPStream* stream, int32_t offset, int32_t len, void* buffer) {
  debugLogLine("NPP_Write");
  return 0;
}

void
NPP_StreamAsFile(NPP instance, NPStream* stream, const char* fname) {
  debugLogLine("NPP_StreamAsFile");
}

void
NPP_Print(NPP instance, NPPrint* platformPrint) {
  debugLogLine("NPP_Print");
}

int setSize(NPP instance, int width, int height)
	{
	NPError err;
	NPObject *obj = 0;
	//debugLog(__LINE__, "setSize: bSize:%d, bVersion:%d", NPNFuncs.size, NPNFuncs.version);

	err = NPNFuncs.getvalue(instance, NPNVPluginElementNPObject, &obj);
	NPIdentifier heightIdentifier	= NPNFuncs.getstringidentifier("height");
	NPIdentifier widthIdentifier	= NPNFuncs.getstringidentifier("width");
	debugLog(__LINE__, "	setSize: PluginElementNPObject:%p, heightId:%p, widthId:%p, err:%d", obj, heightIdentifier, widthIdentifier, err);
	if(err)
		return err;
	if(!obj || !heightIdentifier || !widthIdentifier)
		return NPERR_GENERIC_ERROR;

	bool hasPropertyH	= NPNFuncs.hasproperty(instance, obj, heightIdentifier);
	bool hasMethodH		= NPNFuncs.hasmethod(instance, obj, heightIdentifier);
	bool hasPropertyW	= NPNFuncs.hasproperty(instance, obj, widthIdentifier);
	bool hasMethodW		= NPNFuncs.hasmethod(instance, obj, widthIdentifier);
	debugLog(__LINE__, "	setSize: hasPropertyH:%d hasMethodH:%d hasPropertyW:%d hasMethodW:%d", hasPropertyH, hasMethodH, hasPropertyW, hasMethodW);

	if((!hasPropertyH && !hasMethodH) || (!hasPropertyW && !hasMethodW))
		return NPERR_GENERIC_ERROR;
	NPVariant var;
	bool ok;

#if 1
	ok = NPNFuncs.getproperty(instance, obj, heightIdentifier, &var);
	debugLogVariant(__LINE__, &var, ok, "	setSize: 1 - NPN_GetProperty(height)");
	ok = NPNFuncs.getproperty(instance, obj, widthIdentifier, &var);
	debugLogVariant(__LINE__, &var, ok, "	setSize: 1 - NPN_GetProperty(width)");
#endif

	if(height >= 0)
		{
		INT32_TO_NPVARIANT(height, var);
		ok = NPNFuncs.setproperty(instance, obj, heightIdentifier, &var);
		debugLog(__LINE__, "	setSize: NPN_SetProperty(height:%d) = %s", height, ok ? "OK" : "FAILED");
		if(!ok)
			return NPERR_GENERIC_ERROR;
		}
	if(width >= 0)
		{
		INT32_TO_NPVARIANT(width, var);
		ok = NPNFuncs.setproperty(instance, obj, widthIdentifier, &var);
		debugLog(__LINE__, "	setSize: NPN_SetProperty(width:%d) = %s", width, ok ? "OK" : "FAILED");
		if(!ok)
			return NPERR_GENERIC_ERROR;
		}
#if 1
	ok = NPNFuncs.getproperty(instance, obj, heightIdentifier, &var);
	debugLogVariant(__LINE__, &var, ok, "	setSize: 2 - NPN_GetProperty(height)");
	ok = NPNFuncs.getproperty(instance, obj, widthIdentifier, &var);
	debugLogVariant(__LINE__, &var, ok, "	setSize: 2 - NPN_GetProperty(width)");
#endif
	return err;
	}

void HandleDrawEvent(NpPalmDrawEvent *drawEvent,NPP instance)
{
	unsigned int *screenBuffer = (unsigned int*) drawEvent->dstBuffer;
  // PGContextIface *iface = (PGContextIface *) drawEvent->graphicsContext ;
  // iface->clearRect(5,5,20,20);
  // Draw a blue card - w00t
  // for(int i = 0 ;i < 320*480 ; i++) screenBuffer[i] = 0xaabbccdd ;
	InstanceData *data = (InstanceData *)instance->pdata;
	MyObject *myobj = data ? data->object : 0;
	const unsigned int border = colorAqua;
	int x, y, i, top, bottom, left, right;

	const int screenWidth	= drawEvent->dstRowBytes / 4;
	top		= max(0, drawEvent->srcTop);
	bottom	= drawEvent->srcBottom;
	left	= max(0, drawEvent->dstLeft);
	right	= min(screenWidth, drawEvent->dstRight);

	debugLog(__LINE__, "HandleDrawEvent(sl:%i sr:%i st:%i sb:%i dl:%i dr:%i dt:%i db:%i drb:%i [top:%d left:%d bottom:%d right:%d screenWidth:%d] buf:%p myobj:%p)",
		drawEvent->srcLeft,
		drawEvent->srcRight,
		drawEvent->srcTop,
		drawEvent->srcBottom,
		drawEvent->dstLeft,
		drawEvent->dstRight,
		drawEvent->dstTop,
		drawEvent->dstBottom,
		drawEvent->dstRowBytes,
		top,
		left,
		bottom,
		right,
		screenWidth,
		screenBuffer,
		myobj);

	if (!myobj)
		return;

	int event_x = myobj->event_x;
	int event_y = myobj->event_y;
	int solid_lines = myobj->solid_lines;
	int dy = -event_y;
	const unsigned int colors[] = { colorRed, colorBlue, colorFuchsia, colorYellow, colorGreen };
	unsigned int color;
#if 1
	for(y = top, i = 0; y < bottom; y++, i += screenWidth, dy++)
		{
		int dySquare = dy * dy;
		int g = y % 256;
		int b = ((y / 256) * 0x7F) & 0xFF;
		color = colors[(y / 256) % (sizeof(colors) / sizeof(colors[0]))];		//	use one of the our colors
		color -= (y % 256);		//	every row another color
		if(y == top || y >= (bottom - 1))
			{
			for(x = left; x < right; x++)
				screenBuffer[i + x] = border;
			}
		else
			{
			for(x = left; x < right; x++)
				{
				int dx = x-event_x;
				int dxSquare = dx * dx;
				unsigned int thisColor;

				if((dxSquare + dySquare) < 10)
					thisColor = colorWhite;
				else if(x == 0 || x == (screenWidth - 1))
					thisColor = border;
				else
					{
					if(solid_lines)
						thisColor = color;
					else
						thisColor = RGB(x % 256, g, b);
					}
				screenBuffer[i + x] = thisColor;
				}
			}
		}
#else
  int dst_width = drawEvent->dstRight-drawEvent->dstLeft;
  debugLogLine("base: 6");
  int dst_height = drawEvent->dstBottom-drawEvent->dstTop;
  debugLogLine("base: 7");
  int dst_size = dst_width*dst_height;
  debugLogLine("base: 8");

#if 1
  for (int i = 0 ;i < dst_size; i++) {
    int x = i%320;
    int y = i/320;
    int dx = x-event_x;
    int dy = y-event_y;
    int r;
    int g;
    int b;
    if (dx*dx+dy*dy<10) {
      r = 255;
      g = 255;
      b = 255;
    }
    else {
      r = x%256;
      g = y%256;
      b = 0;
    }
    int a = 255;
    screenBuffer[i] = (a<<24) | (r<<16) | (g<<8) | b;
  }
#endif
#endif
}


int16_t
NPP_HandleEvent(NPP instance, void* aEvent) {
  NpPalmEventType *palmEvent = (NpPalmEventType *) aEvent;
  switch (palmEvent->eventType) {
    case npPalmPenDownEvent:
			debugLog(__LINE__, "npPalmPenDownEvent(x:%d, y:%d, mod:%d)",	palmEvent->data.penEvent.xCoord, palmEvent->data.penEvent.yCoord, palmEvent->data.penEvent.modifiers);
    break;
    case npPalmPenUpEvent:
			debugLog(__LINE__, "npPalmPenUpEvent(x:%d, y:%d, mod:%d)",		palmEvent->data.penEvent.xCoord, palmEvent->data.penEvent.yCoord, palmEvent->data.penEvent.modifiers);
    break;
    case npPalmPenMoveEvent:
		{
		InstanceData *data = (InstanceData *)instance->pdata;
		MyObject *myobj = data->object;
		debugLog(__LINE__, "npPalmPenMoveEvent(x:%d, y:%d, mod:%d, event_x:%d, event_y:%d)",	palmEvent->data.penEvent.xCoord, palmEvent->data.penEvent.yCoord, palmEvent->data.penEvent.modifiers, myobj->event_x, myobj->event_y);
		}
    break;
    case npPalmKeyDownEvent:
			debugLog(__LINE__, "npPalmKeyDownEvent(chr:%d, mod:%d, rawKey:%d, rawMod:%d)",		palmEvent->data.keyEvent.chr, palmEvent->data.keyEvent.modifiers, palmEvent->data.keyEvent.rawkeyCode, palmEvent->data.keyEvent.rawModifier);
    break;
    case npPalmKeyUpEvent:
			debugLog(__LINE__, "npPalmKeyUpEvent(chr:%d, mod:%d, rawKey:%d, rawMod:%d)",		palmEvent->data.keyEvent.chr, palmEvent->data.keyEvent.modifiers, palmEvent->data.keyEvent.rawkeyCode, palmEvent->data.keyEvent.rawModifier);
    break;
    case npPalmKeyRepeatEvent:
			debugLog(__LINE__, "npPalmKeyRepeatEvent(chr:%d, mod:%d, rawKey:%d, rawMod:%d)",	palmEvent->data.keyEvent.chr, palmEvent->data.keyEvent.modifiers, palmEvent->data.keyEvent.rawkeyCode, palmEvent->data.keyEvent.rawModifier);
    break;
    case npPalmKeyPressEvent:
			debugLog(__LINE__, "npPalmKeyPressEvent(chr:%d, mod:%d, rawKey:%d, rawMod:%d)",	palmEvent->data.keyEvent.chr, palmEvent->data.keyEvent.modifiers, palmEvent->data.keyEvent.rawkeyCode, palmEvent->data.keyEvent.rawModifier);
    break;
    case npPalmDrawEvent:
      debugLogLine("NPP_HandleEvent: npPalmDrawEvent");
      HandleDrawEvent(&palmEvent->data.drawEvent,instance);
    break;
    default:
			debugLog(__LINE__, "NPP_HandleEvent: unknown (%d)", palmEvent->eventType);
    break;
  }
  return 1;
}

void
NPP_URLNotify(NPP instance, const char* URL, NPReason reason, void* notifyData) {
  debugLogLine("NPP_URLNotify");
}

const char * varname(NPPVariable variable)
{
  switch (variable) {
    case NPPVpluginNameString:
    return "NPPVpluginNameString";
    break;
    case NPPVpluginDescriptionString:
    return "NPPVpluginDescriptionString";
    break;
    case NPNVnetscapeWindow:
    return "NPNVnetscapeWindow";
    break;
    case NPPVpluginScriptableIID:
    return "NPPVpluginScriptableIID";
    break;
    case NPPVjavascriptPushCallerBool:
    return "NPPVjavascriptPushCallerBool";
    break;
    case NPPVpluginKeepLibraryInMemory:
    return "NPPVpluginKeepLibraryInMemory";
    break;
    case NPPVpluginScriptableNPObject:
    return "NPPVpluginScriptableNPObject";
    break;
    case NPPVpluginNeedsXEmbed:
    return "NPPVpluginNeedsXEmbed";
    break;
    case npPalmEventLoopValue:
    return "npPalmEventLoopValue";
    break;
    case npPalmCachePluginValue:
    return "npPalmCachePluginValue";
    break;
    case npPalmApplicationIdentifier:
    return "npPalmApplicationIdentifier";
    break;
    default:
    return "unknown";
    break;
  }
}

NPError
NPP_GetValue(NPP instance, NPPVariable variable, void *value) {
  logmsgv("NPP_GetValue",variable);
  switch (variable) {
    case NPPVpluginScriptableNPObject:
      *(NPObject **)value = GetScriptableObject(instance);
      break;
    default:
      return NPERR_GENERIC_ERROR;
  }
  return NPERR_NO_ERROR;
}

NPError
NPP_SetValue(NPP instance, NPNVariable variable, void *value) {
   debugLogLine("NPP_SetValue");
  // logmsgv("NPP_SetValue",variable);
  return NPERR_GENERIC_ERROR;
}
